import argparse
import asyncio
import collections
import logging


class Server:
    def __init__(self, host='localhost', port=5000,
                 buffer_chunk_size=10**4, buffer_length_limit=10**4):
        self._host = host
        self._port = port
        self._stopping = False
        self.buffer = collections.deque()  # Shared queue of bytes
        self._buffer_chunk_size = buffer_chunk_size   # How many bytes per chunk
        self._buffer_length_limit = buffer_length_limit  # How many chunks in buffer
        self._working = False
        self.at_eof = False
        self._server = None

    @property
    def host(self) -> str:
        return self._host

    @property
    def port(self) -> int:
        return self._port

    @property
    def stopping(self) -> bool:
        return self._stopping

    @property
    def buffer_length_limit(self) -> int:
        return self._buffer_length_limit

    @property
    def buffer_chunk_size(self) -> int:
        return self._buffer_chunk_size

    @property
    def working(self) -> bool:
        return self._working

    @property
    def server(self) -> asyncio.base_events.Server:
        return self._server

    async def run_reader(self, reader):
        while not self.stopping:
            try:
                # Stop if buffer is full
                while len(self.buffer) >= self.buffer_length_limit:
                    await asyncio.sleep(1)
                    continue
                input_data = await reader.read(self.buffer_chunk_size)
                if reader.at_eof():
                    self.at_eof = True
                self.buffer.append(input_data)
            except Exception as e:
                logging.error(e)

    async def run_writer(self, writer):
        while not self.stopping:
            try:
                # Slow down if buffer is short
                if len(self.buffer) < 3:
                    await asyncio.sleep(.1)
                try:
                    input_data = self.buffer.popleft()
                except IndexError:
                    if not self.at_eof:
                        continue
                    else:
                        writer.write_eof()
                        await writer.drain()
                        self.at_eof = False
                        break
                writer.write(input_data)
                await writer.drain()
            except Exception as e:
                logging.error(e)

    async def connect(self,
                      reader: asyncio.StreamReader,
                      writer: asyncio.StreamWriter):
        """Connect with client.

        Decide whether client is sender or receiver and start transmission.
        """
        peer_is_sender = not self.working  # TODO: ask peer role
        self._working = True
        if peer_is_sender:
            logging.info("Sender is connecting...")
            await self.run_reader(reader=reader)
            logging.info("Incoming transmission ended")
        else:
            logging.info("Receiver is connecting...")
            await self.run_writer(writer=writer)
            logging.info("Outgoing transmission ended")
        return

    async def run_server(self):
        self._server = await asyncio.start_server(
            client_connected_cb=self.connect,
            host=self.host,
            port=self.port
        )
        async with self.server:
            await self.server.serve_forever()
        return

    def stop(self, *_):
        if self.working and not self.stopping:
            logging.info("Received interruption signal, stopping...")
            self._stopping = True
        else:
            raise KeyboardInterrupt("Not working yet...")


if __name__ == '__main__':
    # noinspection SpellCheckingInspection
    log_formatter = logging.Formatter(
        "%(asctime)s [%(module)-15s %(levelname)-8s]     %(message)s",
        style='%'
    )
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(log_formatter)
    console_handler.setLevel(logging.DEBUG)
    root_logger.addHandler(console_handler)

    # Parse command-line arguments
    parser = argparse.ArgumentParser(description='Run server',
                                     allow_abbrev=False)
    parser.add_argument('--_host', type=str,
                        default=None,
                        required=False,
                        help='server address')
    parser.add_argument('--_port', type=int,
                        default=None,
                        required=False,
                        help='server _port')
    args = vars(parser.parse_args())
    host = args['_host']
    port = args['_port']

    # If _host and _port are not provided from command-line, try to import them
    if host is None:
        try:
            from config import host
        except ImportError:
            host = None
    if port is None:
        try:
            from config import port
        except ImportError:
            port = None

    # If import fails, prompt user for _host or _port
    while host is None:
        host = input("Enter _host:\t\t\t\t\t\t")
    while port is None:
        try:
            port = int(input("Enter _port:\t\t\t\t\t\t"))
        except ValueError:
            logging.info("Invalid _port. Enter a valid _port number!")
            port = None

    loop = asyncio.get_event_loop()
    server = Server(
        host=host,
        port=port,
    )
    logging.info("Starting file bridging server...")
    try:
        loop.run_until_complete(server.run_server())
    except KeyboardInterrupt:
        logging.info("Stopping...")
    loop.close()
